import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pylab
import skimage
from scipy import ndimage
from skimage.filters import difference_of_gaussians, gaussian
import pims



'''

Performs bandpass filtering.  Gaussian blurs result to smooth pixel noise to help the contour finding.


Parameters
----------
image: numpy array

low_sigma : int, low frequency cutoff for bandpass filter

blur_sigma: int, radius for gaussian blur (done to aid contour finding after filtering)

Returns
-------
processed_image: numpy array 

'''

def preprocess_image(image, low_sigma = 1, blur_sigma = 3):

    image_result = difference_of_gaussians(image, low_sigma = low_sigma)  #bandpass filter

    image_result = gaussian(image_result, sigma = blur_sigma)  #blur to aid contour finding
    
    return image_result


    '''
    
    Finds contours in the image after preprocessing.

    Input: image (array-like)

    Output: list of contours, each item in the list contains x & y positions of each contour

    Parameters
    ----------
    image: numpy array

    contour_value: float, threshold to use for identifying contours, if not specified it is set 
                to 5*standard_deviation of image values (preprocessing ensures intensities are 
                gaussian centered at zero)


    '''
def locate_streaks(image, contour_value = None):

    processed_image = preprocess_image(image)

    if contour_value == None:
        contour_value = 5*np.std(processed_image)
  
    contours =  skimage.measure.find_contours(processed_image, contour_value) 

    return contours

   


import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import skimage.draw
import networkx as nx
import itertools
import pylab
from scipy.spatial import KDTree
from skimage.morphology import medial_axis
from skimage.filters import difference_of_gaussians, window, gaussian
import matplotlib as mpl
import pandas as pd


#########################################################
'''
centerlines identifies the medial axis of each contour found in an image

helper functions: polycenterline: uses polygon defined by contour coords to find medial axis via skeltoniazation
'''
#########################################################

'''

Uses a contour to find the centerline (medial axis)

Input: contours, list of contours, each item in the list contains x & y positions of each contour
	   image: numpy array, image corresponding to contours list

Output: list of x,y points in the medial axis

'''


def centerlines(contours, image):
	ma = []
	for contour in contours:
		try:
			ctrline = poly_centerline(contour, image)
			srt_ctrline = sort_points(ctrline)
			ma.append(srt_ctrline)
		except:
			ma.append([])

	return ma


'''

poly_centerline: returns the medial axis of the polygon defined by points found in the array image_result.  
					This is a wrapper on skimage's medial axis function, which processes a thresholded image,
					and thus the returned coordinates are ints representing the skeleton (no-subpixel resolution)

inputs: points: numpy array of x,y coordinates defining a polygon, (x_points are in column 1
				and the y points are in column 0)

		image: numpy array that is the image the polygon was found in, needed because a thresholded image
				is needed in this workflow



outputs: coords: numpy array (x=coords[:,0], y = coords[:,1]) of the coordinates of the medial axis
'''
def poly_centerline(points, image):

	#now use the polygon points to define a thresholded image
	thresh = threshold_from_contour(points, image)

	#now find the medial axis (this is a mask on the image)
	cl = medial_axis(thresh)

	#now extract the coordinates from the mask
	coords = np.transpose((cl == True).nonzero())   

	return coords


'''
wrapper around skimage polygon function to return thresholded image as defined by contour points
'''


def threshold_from_contour(points, image):
	rr, cc = skimage.draw.polygon(points[:,0],points[:,1])   
	mask = np.zeros(image.shape) 
	mask[rr,cc]=1   

	return mask




'''
sort_points: creates a graph from points, and then uses the shortest path in the graph to sort the points

inputs: points: numpy array of x,y coordinates defining the medial axis of a polygon 

outputs: coords: numpy array (x=coords[:,0], y = coords[:,1]) of the coordinates of the medial axis but now ordered
'''
def sort_points(points):


	G = nx.Graph()  # A graph to hold the nearest neighbours
	G.add_nodes_from(range(len(points)))  

	tree = KDTree(points, leafsize=2)  # Create a distance tree
	#nn = tree.query(points, k = 3) #identify 2 nearest neighbors (first one is yourself)
	nn = tree.query_pairs(1.5)

	#define edges from nn list
	for item in nn:
		for entry in item[1::]:
			G.add_edge(item[0], entry)

	#identify degree of each node
	degrees = np.array([val for (node, val) in G.degree()])

	free_ends = np.where(degrees < 2)[0]  #dentify nodes with degree < 2, these are candidate source/target nodes
	shortest_paths = []
	source_target_combos = list(itertools.combinations(free_ends,2))  
	for pair in source_target_combos:
		shortest_paths.append(nx.shortest_path(G, source=pair[0], target=pair[1]))

	#find the longest shortest path in case there were stubby ends
	sort_order = max(shortest_paths, key = len)

	sorted_points = points[sort_order]

	return sorted_points









import matplotlib.pyplot as plt
import matplotlib as mpl
import pylab
import numpy as np
import pandas as pd
from skimage import (data, restoration, util)
import skimage
from scipy import ndimage
from skimage.filters import difference_of_gaussians, window, gaussian
import pims
from scipy.optimize import curve_fit
from scipy.special import erf


'''  
    Finds streak attributes

    Input: contours, list of contours, each item in the list contains x & y positions of each contour

    Output: streak_properties, pandas dataframe, lists contour number, area, perimeter, centroid, bbox properties

    Parameters
    ----------

    contour_value: float, threshold to use for identifying contours, if not specified it is set 
                to 5*standard_deviation of image values (preprocessing ensures intensities are 
                gaussian centered at zero)

    frame_number: float, can be used to label frame when batch processing
                '''


def parameters(contours, frame_num = 0):

    property_list = []
    for i, contour in enumerate(contours):

        x = contour[:,1]; y = contour[:,0];  #seperate x & y points

        #check that contour is closed, e.g. it does not interect the image edge, if closed then measure
        if contour[0][0] != contour[-1][0] or contour[0][1] != contour[-1][1]: 
            pass
        else:
            centroid, area, perimeter = contour_shape_specs(x,y)
            corner, width, height = contour_bounding_box(x,y)
            angle = contour_angle(x,y)

            property_list.append([int(i),centroid[0],centroid[1],area, perimeter, corner[0], corner[1], 
                                  width, height, angle,  frame_num])

    properties = pd.DataFrame(data = np.array(property_list), columns = ['streak_id','x','y','area',
                                                                         'perimeter','corner_x', 'corner_y',
                                                                         'bbox_width', 'bbox_height', 'angle','frame'])
    return properties




###################################################################

    '''
    Helper functions to measure contour properties, borrowed heavily from the 'edge' class in ASTiDe

    input: array of x,y points (floats)

    output: contour_shape_specs: centroid, area, perimeter

            contour_bounding_box: bounding box (x,y) corner and width, height

            contour_angle: angle derived from slope fit to all contour points
                            !!this will give nonsense is the streak is not elongated!!!
    '''

def contour_shape_specs(x, y):

    xyxy = (x[:-1] * y[1:] - x[1:] * y[:-1])
    area = 0.5 * np.sum(xyxy)

    perimeter = np.sum(np.sqrt((x[1:] - x[:-1])**2 +
                                   (y[1:] - y[:-1])**2))

    one_sixth_a = 1. / (6. * area)
    x_center = one_sixth_a * np.sum((x[:-1] + x[1:]) * xyxy)
    y_center = one_sixth_a * np.sum((y[:-1] + y[1:]) * xyxy)

    return (x_center, y_center), area, perimeter

def contour_bounding_box(x, y):
    corner = (min(x), min(y))                                                            

    width = max(x)-min(x)
    height = max(y)- min(y)  

    return corner, width, height

def line(x,m,b):
    return m*x + b


def contour_angle(x,y):

        # Fitting a straight line to each edge.
        p0 = [0., 0.]

        p1, s = curve_fit(line, x, y, p0)

        angle = np.arctan(p1[0]) * 180 / np.pi

        return angle




###################################################################################
#final streak fittting functions to get streak width and height, see details below

#fit_shape is the main function
###################################################################################
'''
extract_streaks: extracts piece of original image using streak bbox, should be used for a single frame!
    - if fit fails to converge, streak is rejected
    - if found width or height is more than 20% larger than bounding box, streak is rejected

input: image, numpy array

       df, pandas dataframe of streak properties

       padding: extra distance added to edge of bbox to facilitate fitting

       filepath: string, default is None, 
                  if a filepath is given, images of streaks overlaid with extracted length will be saved


output: df appended with streak width & height
'''
def fit_shape(image, df, padding = 20, pixels_to_average = 2):

    streak_images = []
    widths = []
    heights = []
    slopes = []

    #go through the dataframe row-by-row
    for idx, streak in df.iterrows():

        #extract the bbox of streak with padding
        x0 = int(streak.corner_x - 2*padding); x1 = int(streak.corner_x + streak.bbox_width + 2*padding);
        y0 = int(streak.corner_y - padding); y1 = int(streak.corner_y + streak.bbox_height + padding);
        streak_image = image[y0:y1, x0:x1]

        #rotate iamge
        rotated_streak =  ndimage.rotate(streak_image, streak.angle, reshape=False, mode = 'nearest')

        #find centerline to determine width and height
        xc = int(np.floor(rotated_streak.shape[1] / 2) + 1)
        yc = int(np.floor(rotated_streak.shape[0] / 2) + 1)

        #extract centerline of image for fitting
        width_cut = np.mean(rotated_streak[yc-pixels_to_average : yc+pixels_to_average + 1, :], axis = 0)
        height_cut = np.mean(rotated_streak[:, xc-pixels_to_average : xc+pixels_to_average + 1], axis = 1)      



        #fit for streak width & height
        try:
            w, m = fit_streak_width(width_cut)
        except:
            w = 0
            m = 0
        try:
            h = fit_streak_height(height_cut)
        except:
            h = 0

        #save fits params in array
        widths.append(w); heights.append(h); slopes.append(m);

   #add columns to df in a way that respects indexing and avoids 'Set without copy' warning
    width_series = pd.Series(widths); height_series = pd.Series(heights); slopes_series = pd.Series(slopes);
    width_series.index = df.index; height_series.index = df.index; slopes_series.index = df.index

    df2 = df.copy()

    df2['width'] = width_series
    df2['height'] = height_series
    df2['slope'] = slopes_series

    #do a course filter to remove bad streaks, i.e ones where fit didn't converge or ones which
    #are more than 20% larger than bbox
    df3 = df2[df2.height != 0]
    df4 = df3[df3.width != 0]
    df5 = df4[df4.width < 1.2*df4.bbox_width]
    df6 = df5[df5.height < 1.2*df5.bbox_height]

    return df6

###################################################################
    
#    Helper functions to fit streaks, see details below

###################################################################


'''
convolved erf function for fitting streak width
parameters: 

amp = amplitude
w0: width of bump
L:
s:
m: slope of offset line
'''
def erf_fit(x,amp,w0,L,s,m,b,a):
    erf_convolve = amp*(erf((x-w0)/s/(2**(1/4)))-erf((x-w0-L)/s/(2**(1/4))))/2  
    return erf_convolve*(1-m*x)+b*x+a


'''
gauss function for fitting streak height
parameters: 

amp = amplitude
offset: offset of gaussian center from zero
sigma: width of the bump
m: slope of offset line
'''
def gauss(x,amp,offset,sigma,b,a):
    ans = amp*np.exp(-(x-offset)**2/(2*sigma**2))
    return ans+b*x+a


'''helper function, fits for streak width using erf_convolve
input: centerline: numpy array, bump-like shape assumed

output: w, streak width (width of convolved erf)
'''

def fit_streak_width(centerline):
    #fit for the streak width
    yy = centerline
    xx=np.arange(0,len(yy))

    #normalize 
    yy=(yy-np.min(yy))/(np.max(yy)-np.min(yy))

    #initial guess
    
    #amp,w0,L,s,m,a,b
    amp=1
    w0=20
    L=len(yy)-40
    s=20
    m=(yy[-1]-yy[0])/(xx[-1]-xx[0])
#    a=(yy[-1]-yy[0])/(xx[-1]-xx[0])
    b=0.01##np.mean(yy[0:5])
    a=0.01
    p0=[amp,w0,L,s,m,b,a]

    pred_params, uncert_cov = curve_fit(erf_fit, xx, yy, p0=p0,method='lm')

    # plt.figure()
    # plt.plot(xx,yy,marker='o',c='k',ls='None',ms=3)
    # plt.plot(xx,erf_fit(xx,*pred_params),lw=2,c='r')

    w = pred_params[2]

    return abs(w), pred_params[4]  #there is a degeneracy where sigma can be negative

'''helper function, fits for streak width using erf_convolve
input: centerline: numpy array, bump-like shape assumed

output: h, streak height (width of fit gaussian)
'''

def fit_streak_height(centerline):
             #fit for the streak height
        yy = centerline
        xx = np.arange(0,len(yy))

        #normalize 
        yy=(yy-np.min(yy))/(np.max(yy)-np.min(yy))

        #initial guess
        p0=[1,10,(len(yy)-20),0.01,0.01]

        pred_params, uncert_cov = curve_fit(gauss, xx, yy, p0=p0,method='lm')

        # plt.figure()
        # plt.plot(xx,yy,marker='o',c='k',ls='None',ms=3)
        # plt.plot(xx,gauss(xx,*pred_params),lw=2,c='r')

        h = pred_params[2]

        return abs(h)  #there is a degeneracy where sigma can be negative




import matplotlib.pyplot as plt
import matplotlib as mpl
import pylab
import pandas as pd
import numpy as np



###################################################################################
'''
plot_contours - to be used to plot contours, for centerlines see plot_centerlines

inputs:
    im: image of streaks, numpy array
    contours: list of contours, each list entry is numpy array of x,y positions in contour

output: None, displays image
'''
###################################################################################

def plot_streaks(contours, im):
    fig, ax = plt.subplots()
    ax.imshow(im, cmap = 'Greys_r')

    for i, contour in enumerate(contours):
        ax.plot(contour[:,1], contour[:,0], linewidth=2)
        ax.text(contour[0][1], contour[0][0], str(int(i)), fontsize = 12, c = 'y')
    plt.show()

###################################################################################
'''
plot_contours_label - to be used to plot contours that have been characterized by running parameters

inputs:
    df: pandas dataframe with streak information (output of measure.parameters)
    im: image of streaks, numpy array
    contours: list of contours, each list entry is numpy array of x,y positions in contour

output: None, displays image
'''
###################################################################################

def plot_filtered_streaks(contours, df, im):
    fig, ax = plt.subplots()
    ax.imshow(im, cmap = 'Greys_r')

    for ind in df.streak_id.unique():

        contour = contours[int(ind)]
        ax.plot(contour[:,1], contour[:,0], linewidth=2, )
        ax.text(contour[0][1], contour[0][0], str(int(ind)), fontsize = 12, c = 'y')
        
    plt.show()

###################################################################################
'''
plot_fit - plots width and length identified for straight streaks

inputs:
    df: pandas dataframe with streak information
    im: image of streaks, numpy array
    contours: list of contours, each list entry is numpy array of x,y positions in contour

output: None, displays image
'''
###################################################################################

def plot_fit(im, df):

    fig, ax = plt.subplots()
    ax.imshow(im, cmap = 'Greys_r')
    
    for idx, row in df.iterrows():
        cx = row.x; cy = row.y;  #centroid
        theta = -row.angle*np.pi/180 #rotation angle
        #plot width line
        lx, ly = rotate(cx - row.width/2, cy, cx, cy, theta)
        rx, ry = rotate(cx + row.width/2, cy, cx, cy, theta)
        plt.plot([lx,rx],[ly,ry],c='r')
        #plot height line
        lx, ly = rotate(cx, cy - row.height, cx, cy, theta)
        rx, ry = rotate(cx, cy + row.height, cx, cy, theta)
        plt.plot([lx,rx],[ly,ry],c='r')
        ax.text(row.x, row.y, str(int(row.streak_id)), fontsize = 12,c='y')

    plt.show()


#helper function for plot_fit
def rotate(x,y,cx,cy,theta):  #theta should be in radians
    xp = ((x-cx)*np.cos(theta) + (y - cy)*np.sin(theta)) + cx
    yp = (-(x-cx)*np.sin(theta) + (y - cy)*np.cos(theta)) + cy

    return xp, yp



###################################################################################
#plot_centerline - plots the centerline found for curvy streaks

'''
inputs:
    cl: list of centerlines, list of numpy arrays
    im: image of streaks, numpy array
    show_point_order: plot centerline points using a colormap to indicate the ordering of points, dafualt is True
    cmap: colormap for centerline to indicate point order, default is rainbow
    color: color for streaks if you do not want to indicate point order, default is red

output: None, displays image
'''
###################################################################################


def plot_centerlines(cl, image, show_point_order = False, color = 'r', cmap = 'rainbow'):
    fig, ax = plt.subplots()
    plt.imshow(image, cmap = 'Greys_r')

    for i, item in enumerate(cl):
        if len(item) > 0:
            if show_point_order:
                ax.scatter(item[:,1],item[:,0],s=5,marker='o',c=np.arange(len(item)), cmap = cmap)  
                ax.text(item[0][1], item[0][0], str(int(i)), fontsize = 12, c = 'y')
            else:
                ax.plot(item[:,1],item[:,0],c=color)
                ax.text(item[0][1], item[0][0], str(int(i)), fontsize = 12, c = 'y')
    plt.show()








import numpy as np

'''  
    Course first-pass filter to elimate clearly mis-sized streaks as well as those that are too close to the edge.

    Filters found streaks, defaults choose longer, narrower streaks above a certain size threshold;
    streaks are also filtered to remove those near edges

    Input: streaks, pandas dataframe of streaks generated by measure (lists contour number, area, perimeter, centroid, bbox properties)

    Output: final_filtered, pandas dataframe of filtered streaks

    Parameters
    ----------

    min_height: minimum bbox height of contours to use for filtering, default is 7

    min_aspect: minimum aspect ratio (bbox width / bbox height) of contours to use for filtering, default is 2

    remove_edges: boolean, if True (default) removes contours whose centroid is within edge_filter of the image border

    edge_filter: size in pixels of excluded region for edge filtering, default is 100

    image_size: tuple (width, height), size of image to use for edge filtering
                '''


def filter(streaks, min_height = 7, min_aspect = 2, remove_edges = True, edge_filter = 100, image_size = (2048, 2048)):
    height_filter = streaks[streaks.bbox_height > min_height]
    aspect_filter = height_filter[(height_filter.bbox_width / height_filter.bbox_height) > 2]

    if remove_edges:
        high_x = aspect_filter[aspect_filter.x > edge_filter]
        center_x = high_x[high_x.x < (image_size[0] - edge_filter)]

        high_y = center_x[center_x.y > edge_filter]
        center_y = high_y[high_y.y < (image_size[1] - edge_filter)]

        edge_filtered = center_y

    else:
        edge_filtered = aspect_filter

    final_filtered = edge_filtered

    return final_filtered



'''
Basic contour filtering, removes only contours which are open (i.e intersect with the edge) and
contours which are too small, below a minimum diagonal size

This should be used if your contours are not rectangles, and thus the bounding box filtering will not work 
well.  Note that this function filters contours, not the dataframe of measurements

Input: Input: contours, list of contours, each item in the list contains x & y positions of each contour

Output: list of filtered contours, each item in the list contains x & y positions of each contour

Parameters
----------
min_diag_size: float, diagonal size threshold to use for filtering out small contours, 
'''

def filter_diag_size(contours, min_diag_size = 80):
    filtered = []
    for i, contour in enumerate(contours):

        x = contour[:,1]; y = contour[:,0];  #seperate x & y points

        #check that contour is closed, e.g. it does not interect the image edge, if closed then measure
        if contour[0][0] != contour[-1][0] or contour[0][1] != contour[-1][1]: 
            pass
        else:
            width = max(x)-min(x)
            height = max(y)- min(y) 
            diag_size = np.sqrt(width**2 + height**2)
            if diag_size > min_diag_size:
                filtered.append(contour)

    return filtered



import setuptools
from setuptools import setup

setup(
   name='fsvpy',
   version='1.0.1',
   description='streak finding module',
   long_description='Efficient algorithm for generalized streak finding for velocimetry applications.',
   author='Michelle Driscoll',
   author_email='michelle.driscoll@northwestern.edu',
   url="",
   packages = setuptools.find_packages(),
   install_requires=['pims', 'matplotlib', 'pandas', 'numpy', 'scikit-image', 'networkx'] #external packages as dependencies
)
